<?php

namespace HaoZiTeam\WPChatGPT\Service;

defined( 'ABSPATH' ) || exit;

use Exception;
use HaoZiTeam\ChatGPT\V2;
use GuzzleHttp\Psr7;
use WP_REST_Request;
use WP_REST_Response;
use WP_REST_Server;

/**
 * Class Chat
 * 实现插件的对话接口
 * @package HaoZiTeam\ChatGPT\Service
 */
class Chat {

	private $settings;

	public function __construct() {
		$this->settings = get_option( 'wp-chatgpt' );

		register_rest_route( 'wp-chatgpt', 'user-info', array(
			'methods'             => WP_REST_Server::READABLE,
			'callback'            => array( $this, 'user_info' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'message-process', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'message_process' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'delete-message', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'delete_message' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'conversation-list', array(
			'methods'             => WP_REST_Server::READABLE,
			'callback'            => array( $this, 'conversation_list' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'new-conversation', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'new_conversation' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'update-conversation', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'update_conversation' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'delete-conversation', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'delete_conversation' ),
			'permission_callback' => '__return_true',
		) );
		register_rest_route( 'wp-chatgpt', 'clear-conversation', array(
			'methods'             => WP_REST_Server::CREATABLE,
			'callback'            => array( $this, 'clear_conversation' ),
			'permission_callback' => '__return_true',
		) );
	}

	/**
	 * 获取用户信息
	 *
	 * @return WP_REST_Response
	 */
	public function user_info(): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		$user = wp_get_current_user();

		return $this->success( '获取成功', [
			'name'        => $user->display_name,
			'avatar'      => get_avatar_url( $user->ID ),
			'description' => $user->description,
		] );
	}

	/**
	 * 聊天流式接口
	 * 通过缓冲区刷新的方式实现聊天
	 *
	 * @param WP_REST_Request $request
	 *
	 * @return WP_REST_Response
	 */
	public function message_process( WP_REST_Request $request ): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		$options = $request->get_param( 'options' );

		if ( ! isset( $options['conversationId'] ) ) {
			return $this->error( '会话信息不存在' );
		}

		if ( empty( $request->get_param( 'prompt' ) ) ) {
			return $this->error( '请输入内容' );
		}
		// 判断敏感词
		if ( ! empty( $this->settings['sensitive-words'] ) ) {
			foreach ( $this->settings['sensitive-words'] as $word ) {
				if ( ! empty( $word ) && str_contains( $request->get_param( 'prompt' ), $word['sensitive-word'] ?? null ) ) {
					// 判断敏感词是否大于3个字符，如果大于3个字符则只显示第一个和最后一个字符
					if ( mb_strlen( $word['sensitive-word'] ) > 3 ) {
						$word['sensitive-word'] = mb_substr( $word['sensitive-word'], 0, 1 ) . str_repeat( '*', mb_strlen( $word['sensitive-word'] ) - 2 ) . mb_substr( $word['sensitive-word'], - 1 );
					}

					return $this->error( '您输入的内容包含敏感词：' . $word['sensitive-word'] );
				}
			}
		}

		// 尝试获取会话评论
		$conversation = get_post( $options['conversationId'] );
		if ( ! $conversation ) {
			return $this->error( '会话不存在' );
		}

		// 检查是否是当前用户的会话
		if ( $conversation->post_author != get_current_user_id() ) {
			return $this->error( '无权访问' );
		}

		// 获取评论
		$comments = get_comments( [
			'post_id' => $conversation->ID,
		] );

		/**
		 * 检查是否符合要求
		 * 1. 评论为空，且有 parentMessageId 且 parentMessageId 不为 0 （首条消息 parentMessageId 为 0）
		 * 2. 评论不为空，且有 regenerate 但是没有 parentMessageId
		 */
		if ( empty( $comments ) && isset( $options['parentMessageId'] ) && (int) $options['parentMessageId'] != 0 ) {
			return $this->error( '消息错误，请刷新页面，若仍报错请删除重新开启会话' );
		}
		if ( ! empty( $comments ) && ! empty( $options['regenerate'] ) && empty( $options['parentMessageId'] ) ) {
			return $this->error( '消息错误，请刷新页面，若仍报错请删除重新开启会话' );
		}
		$options['parentMessageId'] = $options['parentMessageId'] ?? 0;

		// 倒序循环评论作为聊天记录
		$text_count = 0;
		$count      = count( $comments );
		// 为首次聊天自动设置标题
		if ( $count == 0 ) {
			wp_update_post( [
				'ID'         => $conversation->ID,
				'post_title' => $request->get_param( 'prompt' ),
			] );
		}

		// 取出最后一条消息作为父消息
		// 如果有 regenerate 则使用 parentMessageId 作为父消息
		$parent_message_id = $comments[0]->comment_ID ?? $options['parentMessageId'];
		if ( ! empty( $options['regenerate'] ) ) {
			$parent_message_id = $options['parentMessageId'] ?? 0;
		}

		// 随机选择账号
		$accounts      = $this->settings['openai-accounts'] ?? [];
		$account_count = count( $accounts );
		if ( $account_count == 0 ) {
			return $this->error( '请先配置 AI 账号' );
		}
		$index   = rand( 1, $account_count );
		$account = $accounts[ $index - 1 ];

		$chat_gpt      = new V2( $account['account-secret-key'] ?? null, $this->settings['openai-api-server'] ?? null );
		$system_prompt = empty( $this->settings['chatgpt-system-prompt'] ) ? '你是 ChatGPT，一个由 OpenAI 训练的大型语言模型，模型是 GPT-3.5，当前时间是 {datetime}。尽可能简洁地回答。' : $this->settings['chatgpt-system-prompt'];
		// 尝试替换系统提示中的变量
		$system_prompt = str_replace( '{datetime}', date( 'Y-m-d H:i:s' ), $system_prompt );

		$chat_gpt->addMessage( $system_prompt, 'system' );

		for ( $i = $count - 1; $i >= 0; $i -- ) {
			$comment = $comments[ $i ];
			// 判断一下$comment->comment_ID是否大于等于$options['parentMessageId'] + 1
			// 这是为了在重载时，不添加新的评论
			// +1 是因为 parentMessageId 是最后一条问题的ID，而不是最后一条回答的ID
			if ( isset( $options['parentMessageId'] ) && $comment->comment_ID >= $options['parentMessageId'] + 1 ) {
				// 如果大于等于，跳过
				continue;
			}
			// 取评论的meta
			$inversion = get_comment_meta( $comment->comment_ID, 'inversion', true );// yes: 用户提问，no: ChatGPT 回答
			if ( ! empty( $comment->comment_content ) ) {
				$chat_gpt->addMessage( $comment->comment_content, $inversion === 'yes' ? 'user' : 'assistant' );
				// 统计content的字数
				$text_count += mb_strlen( $comment->comment_content );
			}
			if ( $text_count > 1000 ) {
				// 超过1000字，不再添加
				break;
			}
		}

		try {
			$answer = $chat_gpt->ask( $request->get_param( 'prompt' ), null, true );
		} catch ( Exception $e ) {
			$error = $e->getMessage();
			if ( $this->settings['openai-api-server'] ?? false ) {
				$error = str_replace( $this->settings['openai-api-server'], '****.****.****', $error );
			}
			$error .= "\n";
			$error .= 'Account: ' . $account['account-note'] ?? '未知';

			return $this->error( $error );
		}

		// 为重载，不保存提问
		if ( empty( $options['regenerate'] ) ) {
			global $current_user;
			$question_id = wp_insert_comment( [
				'comment_post_ID' => $conversation->ID,
				'comment_author'  => $current_user->display_name,
				'comment_content' => $request->get_param( 'prompt' ),
				'user_id'         => $current_user->ID,
				'comment_meta'    => [
					'parent_message_id' => $parent_message_id,
					'inversion'         => 'yes',
					'prompt'            => $request->get_param( 'prompt' ),
				],
			] );
		} else {
			$question_id = $options['parentMessageId'];
		}

		// 保存回答
		if ( empty( $options['regenerate'] ) ) {
			global $current_user;
			$answer_id = wp_insert_comment( [
				'comment_post_ID' => $conversation->ID,
				'comment_author'  => 'AI',
				'comment_content' => '',
				'user_id'         => $current_user->ID,
				'comment_meta'    => [
					'parent_message_id' => $question_id,
					'inversion'         => 'no',
					'prompt'            => $request->get_param( 'prompt' ),
				],
			] );
		} else {
			wp_update_comment( [
				'comment_ID'      => $options['regenerate'],
				'comment_content' => '',
			] );
			$answer_id = $options['regenerate'];
		}

		header( 'Content-type: application/octet-stream' );
		header( 'Cache-Control: no-cache' );
		header( 'X-Accel-Buffering: no' );
		header( 'Access-Control-Allow-Origin: *' );

		ob_end_clean();
		$temp       = '';
		$message_id = '';
		$first      = true;
		while ( ! $answer->eof() ) {
			$raw  = Psr7\Utils::readLine( $answer );
			$line = $chat_gpt->formatStreamMessage( $raw );
			if ( $chat_gpt->checkStreamFields( $line ) ) {
				if ( ! $first ) {
					echo "\n";
				}
				$first      = false;
				$temp       .= $line['choices'][0]['delta']['content'];
				$single     = $line['choices'][0]['delta']['content'];
				$message_id = $line['message_id'];

				echo json_encode( [
					"role"            => "assistant",
					"id"              => $answer_id,
					'conversationId'  => $conversation->ID,
					"parentMessageId" => $question_id,
					"text"            => $temp,
					"delta"           => $single,
				], JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_LINE_TERMINATORS );
				// 判断敏感词
				if ( ! empty( $this->settings['sensitive-words'] ) ) {
					foreach ( $this->settings['sensitive-words'] as $word ) {
						if ( ! empty( $word ) && str_contains( $temp, $word['sensitive-word'] ?? null ) ) {
							return $this->error( '返回的内容包含敏感词' );
						}
					}
				}
			}
			unset( $raw, $line );
			ob_flush();
			flush();
		}

		// 更新回答meta
		update_comment_meta( $answer_id, 'message_id', $message_id );
		wp_update_comment( [
			'comment_ID'      => $answer_id,
			'comment_content' => $temp,
		] );

		exit;
	}

	/**
	 * 删除消息
	 *
	 * @param WP_REST_Request $request
	 *
	 * @return WP_REST_Response
	 */
	public function delete_message( WP_REST_Request $request ): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 获取参数
		$conversation_id = $request->get_param( 'uuid' );
		$message_id      = $request->get_param( 'message' );

		// 获取会话文章
		$post = get_post( $conversation_id );
		if ( ! $post ) {
			return $this->error( '会话不存在' );
		}

		// 检查是否是当前用户的会话
		if ( (int) $post->post_author != get_current_user_id() ) {
			return $this->error( '无权删除' );
		}

		// 获取评论
		$comment = get_comment( $message_id );
		if ( ! $comment ) {
			return $this->error( '消息不存在' );
		}

		// 检查是否是当前用户的评论
		if ( (int) $comment->user_id != get_current_user_id() ) {
			return $this->error( '无权删除' );
		}

		// 删除评论
		if ( $this->settings['message-trash'] ?? false ) {
			wp_trash_comment( $comment->comment_ID );
		} else {
			wp_delete_comment( $comment->comment_ID, true );
		}

		// 返回数据
		return $this->success( '删除成功' );
	}

	/**
	 * 获取会话列表
	 *
	 * @return WP_REST_Response
	 */
	public function conversation_list(): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 移除过滤器
		kses_remove_filters();

		// 获取当前用户的会话文章
		$posts = get_posts( [
			'post_type'      => 'cg_conversations',
			'posts_per_page' => - 1,
			'author'         => get_current_user_id(),
		] );

		// 如果没有会话，新建一个
		if ( empty( $posts ) ) {
			wp_insert_post( [
				'post_title'  => "New Chat",
				'post_type'   => 'cg_conversations',
				'post_status' => 'publish',
				'post_author' => get_current_user_id(),
			] );
			$posts = get_posts( [
				'post_type'      => 'cg_conversations',
				'posts_per_page' => - 1,
				'author'         => get_current_user_id(),
			] );
		}

		// 取每个会话的所有评论
		$messages = [];
		foreach ( $posts as $post ) {
			$comments = get_comments( [
				'post_id' => $post->ID,
				'order'   => 'ASC',
			] );
			foreach ( $comments as $comment ) {
				$parent_message_id = get_comment_meta( $comment->comment_ID, 'parent_message_id', true );
				$inversion         = get_comment_meta( $comment->comment_ID, 'inversion', true );
				$prompt            = get_comment_meta( $comment->comment_ID, 'prompt', true );

				$messages[ $post->ID ][] = [
					'conversationOptions' => [
						'conversationId'  => $comment->comment_post_ID,
						'parentMessageId' => $parent_message_id,
					],
					'id'                  => $comment->comment_ID,
					'dateTime'            => $comment->comment_date,
					'error'               => false,
					'inversion'           => $inversion === 'yes',
					'text'                => $comment->comment_content,
					'requestOptions'      => [
						'options' => [
							'conversationId'  => $comment->comment_post_ID,
							'parentMessageId' => $parent_message_id,
						],
						'prompt'  => $prompt, // 用户提问的内容
					],
				];
			}
		}

		// 构建返回数据
		$data = [];
		foreach ( $posts as $post ) {
			// 获取文章的 is_edit 字段
			$is_edit           = get_post_meta( $post->ID, 'is_edit', true );
			$data['history'][] = [
				'uuid'   => $post->ID,
				'title'  => $post->post_title,
				'isEdit' => ! empty( $is_edit ),
			];
			$data['chat'][]    = [
				'uuid' => $post->ID,
				'data' => $messages[ $post->ID ] ?? [],
			];
		}

		// 返回数据
		return $this->success( '获取成功', $data );
	}

	/**
	 * 新会话
	 *
	 * @return WP_REST_Response
	 */
	public function new_conversation(): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 移除过滤器
		kses_remove_filters();

		// 创建文章
		$post_id = wp_insert_post( [
			'post_title'  => "New Chat",
			'post_type'   => 'cg_conversations',
			'post_status' => 'publish',
			'post_author' => get_current_user_id(),
		] );

		// 返回数据
		return $this->success( '创建成功', [
			'uuid' => $post_id,
		] );
	}

	/**
	 * 更新会话
	 *
	 * @param WP_REST_Request $request
	 *
	 * @return WP_REST_Response
	 */
	public function update_conversation( WP_REST_Request $request ): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 获取参数
		$conversation_id = $request->get_param( 'uuid' );
		$title           = $request->get_param( 'title' );

		// 获取会话文章
		$post = get_post( $conversation_id );
		if ( ! $post ) {
			return $this->error( '会话不存在' );
		}

		// 检查是否是当前用户的会话
		if ( $post->post_author != get_current_user_id() ) {
			return $this->error( '无权修改' );
		}

		// 更新文章
		wp_update_post( [
			'ID'         => $conversation_id,
			'post_title' => $title,
		] );

		// 返回数据
		return $this->success( '修改成功' );
	}

	/**
	 * 删除会话
	 *
	 * @param WP_REST_Request $request
	 *
	 * @return WP_REST_Response
	 */
	public function delete_conversation( WP_REST_Request $request ): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 获取参数
		$conversation_id = $request->get_param( 'uuid' );

		// 获取会话文章
		$post = get_post( $conversation_id );
		if ( ! $post ) {
			return $this->error( '会话不存在' );
		}

		// 检查是否是当前用户的会话
		if ( $post->post_author != get_current_user_id() ) {
			return $this->error( '无权删除' );
		}

		// 删除文章
		if ( $this->settings['conversation-trash'] ?? false ) {
			wp_trash_post( $post->ID );
		} else {
			wp_delete_post( $post->ID, true );
		}

		// 返回数据
		return $this->success( '删除成功' );
	}

	/**
	 * 清空会话消息
	 *
	 * @param WP_REST_Request $request
	 *
	 * @return WP_REST_Response
	 */
	public function clear_conversation( WP_REST_Request $request ): WP_REST_Response {
		// 检查是否登录
		if ( ! is_user_logged_in() ) {
			return $this->error( '请先登录' );
		}

		// 获取参数
		$conversation_id = $request->get_param( 'uuid' );

		// 获取会话文章
		$post = get_post( $conversation_id );
		if ( ! $post ) {
			return $this->error( '会话不存在' );
		}

		// 检查是否是当前用户的会话
		if ( $post->post_author != get_current_user_id() ) {
			return $this->error( '无权删除' );
		}

		// 获取评论
		$comments = get_comments( [
			'post_id' => $post->ID,
		] );

		// 删除评论
		foreach ( $comments as $comment ) {
			if ( $this->settings['message-trash'] ?? false ) {
				wp_trash_comment( $comment->comment_ID );
			} else {
				wp_delete_comment( $comment->comment_ID, true );
			}
		}

		// 返回数据
		return $this->success( '清空成功' );
	}

	public function success( $message = 'success', $data = null ): WP_REST_Response {
		return new WP_REST_Response( array( 'message' => $message, 'data' => $data, 'status' => 'Success' ) );
	}

	public function error( $message = 'error', $data = null ): WP_REST_Response {
		return new WP_REST_Response( array( 'message' => $message, 'data' => $data, 'status' => 'Failed' ) );
	}

}
